//
//  BarcodePickerController2.m
//  ScanditBarcodeScanner
//
//  Created by Moritz Hartmeier on 20/08/15.
//  Copyright (c) 2015 Scandit AG. All rights reserved.
//

#import "BarcodePickerController2.h"

#import "BarcodeResult+Private.h"
#import "RedLaserSettingsChange.h"


@interface SBSBarcodePickerBase ()
@property (nonatomic, strong) SBSScanSettings* scanSettings;
@end


@interface BarcodePickerController2 () <SBSScanDelegate, SBSOverlayControllerDidCancelDelegate>
@property (nonatomic, strong) NSMutableArray *settingsChanges;
@property (nonatomic) BOOL applyingSettingsChanges;
#if OS_OBJECT_HAVE_OBJC_SUPPORT == 1
@property (nonatomic, strong) dispatch_queue_t settingsChangesQueue;
#else
@property (nonatomic) dispatch_queue_t settingsChangesQueue;
#endif
@end

@implementation BarcodePickerController2
@synthesize delegate = _delegate;
@synthesize useFrontCamera = _useFrontCamera;
@synthesize isFocusing = _isFocusing;
@synthesize exposureLock = _exposureLock;

- (instancetype)init {
    if (self = [self initWithNibName:nil bundle:nil]) {
        
    }
    return self;
}

- (id)initWithCoder:(NSCoder *)aDecoder {
    if (self = [self initWithNibName:nil bundle:nil]) {
        
    }
    return self;
}

- (instancetype)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil {
    if (self = [super initWithSettings:[SBSScanSettings defaultSettings]]) {
        self.settingsChanges = [NSMutableArray array];
        self.applyingSettingsChanges = NO;
        self.settingsChangesQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0);
        
        // Set delegates at the end as accessing the overlay controller will instantiate the view
        // of the picker which call onViewDidLoad which might want to set properties here.
        self.scanDelegate = self;
        self.overlayController.cancelDelegate = self;
    }
    return self;
}

- (void)viewDidAppear:(BOOL)animated {
    [super viewDidAppear:animated];
    
    if ([self isModal]) {
        [self.overlayController showToolBar:YES];
    }
    
    [self startScanning];
}

- (BOOL)isModal {
    return (self.presentingViewController.presentedViewController == self
            || (self.navigationController != nil
                && self.navigationController.presentingViewController.presentedViewController == self.navigationController)
            || [self.tabBarController.presentingViewController isKindOfClass:[UITabBarController class]]);
}

- (void)setUseFrontCamera:(BOOL)useFrontCamera {
    _useFrontCamera = useFrontCamera;
    
    if (useFrontCamera) {
        [self changeToCameraFacing:SBSCameraFacingDirectionFront];
    } else {
        [self changeToCameraFacing:SBSCameraFacingDirectionBack];
    }
}

- (void)doneScanning {
    [self stopScanning];
    
    [self.delegate barcodePickerController:self returnResults:[NSSet set]];
}


#pragma mark - SBSScanDelegate

- (void)barcodePicker:(SBSBarcodePicker *)picker didScan:(SBSScanSession *)session {
    NSSet *allSet = [self setOfBarcodeResultsInArray:session.allRecognizedCodes];
    
    dispatch_sync(dispatch_get_main_queue(), ^{
        if ([self respondsToSelector:@selector(statusUpdated:)]) {
            NSSet *newSet = [self setOfBarcodeResultsInArray:session.newlyRecognizedCodes];
            
            NSMutableDictionary *status = [NSMutableDictionary dictionary];
            [status setObject:allSet forKey:@"FoundBarcodes"];
            [status setObject:newSet forKey:@"NewFoundBarcodes"];
            [status setObject:[NSNumber numberWithBool:([allSet count] > 0) ? YES : NO]
                       forKey:@"Valid"];
            [status setObject:[NSNumber numberWithBool:NO] forKey:@"InRange"];
            [status setObject:[NSNumber numberWithInt:0] forKey:@"Guidance"];
            [status setObject:@"" forKey:@"PartialBarcode"];
            [status setObject:[NSNumber numberWithBool:YES] forKey:@"CameraSupportsTapToFocus"];
            [self statusUpdated:status];

        } else {
            if (self.delegate && [self.delegate respondsToSelector:@selector(barcodePickerController:returnResults:)]) {
                [self.delegate barcodePickerController:self returnResults:allSet];
            }
        }
    });
}

- (NSSet *)setOfBarcodeResultsInArray:(NSArray *)array {
    NSMutableSet *set = [NSMutableSet set];
    for (SBSCode *code in array) {
        BarcodeResult *result = [[BarcodeResult alloc] init];
        if (code.symbology == SBSSymbologyUPCE) {
            result.barcodeType = kBarcodeTypeUPCE;
        } else if (code.symbology == SBSSymbologyEAN8) {
            result.barcodeType = kBarcodeTypeEAN8;
        } else if (code.symbology == SBSSymbologyEAN13 || code.symbology == SBSSymbologyUPC12) {
            result.barcodeType = kBarcodeTypeEAN13;
        } else if (code.symbology == SBSSymbologyQR) {
            result.barcodeType = kBarcodeTypeQRCODE;
        } else if (code.symbology == SBSSymbologyCode128) {
            result.barcodeType = kBarcodeTypeCODE128;
        } else if (code.symbology == SBSSymbologyCode39) {
            result.barcodeType = kBarcodeTypeCODE39;
        } else if (code.symbology == SBSSymbologyDatamatrix) {
            result.barcodeType = kBarcodeTypeDATAMATRIX;
        } else if (code.symbology == SBSSymbologyITF) {
            result.barcodeType = kBarcodeTypeITF;
        } else if (code.symbology == SBSSymbologyFiveDigitAddOn) {
            result.barcodeType = kBarcodeTypeEAN5;
        } else if (code.symbology == SBSSymbologyTwoDigitAddOn) {
            result.barcodeType = kBarcodeTypeEAN2;
        } else if (code.symbology == SBSSymbologyCodabar) {
            result.barcodeType = kBarcodeTypeCodabar;
        } else if (code.symbology == SBSSymbologyCode93) {
            result.barcodeType = kBarcodeTypeCODE93;
        } else if (code.symbology == SBSSymbologyPDF417) {
            result.barcodeType = kBarcodeTypePDF417;
        } else if (code.symbology == SBSSymbologyGS1Databar) {
            result.barcodeType = kBarcodeTypeGS1Databar;
        } else if (code.symbology == SBSSymbologyGS1DatabarExpanded) {
            result.barcodeType = kBarcodeTypeGS1DatabarExpanded;
        }
        
        if (result.barcodeType > 0) {
            if (code.symbology == SBSSymbologyUPC12) {
                result.barcodeString = [NSString stringWithFormat:@"0%@", code.data];
                result.extendedBarcodeString = result.barcodeString;
            } else {
                result.barcodeString = code.data;
                result.extendedBarcodeString = code.data;
            }
            result.associatedBarcode = nil;
            
            result.firstScanTime = [NSDate date];
            result.mostRecentScanTime = [NSDate date];
            result.barcodeLocation = [NSMutableArray array];
            
            [set addObject:result];
        }
    }

    return set;
}


#pragma mark - SBSOverlayControllerDidCancelDelegate

- (void)overlayController:(SBSOverlayController *)overlayController
      didCancelWithStatus:(NSDictionary *)status {
    [self stopScanning];
    [self.delegate barcodePickerController:self returnResults:[NSSet set]];
}


#pragma mark - Symbologies

- (BOOL)scanUPCE {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyUPCE] enabled];
}

- (void)setScanUPCE:(BOOL)scanUPCE {
    [self internalSetSymbology:SBSSymbologyUPCE enabled:scanUPCE];
}

- (BOOL)scanEAN8 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyEAN8] enabled];
}

- (void)setScanEAN8:(BOOL)scanEAN8 {
    [self internalSetSymbology:SBSSymbologyEAN8 enabled:scanEAN8];
}

- (BOOL)scanEAN13 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyEAN13] enabled];
}

- (void)setScanEAN13:(BOOL)scanEAN13 {
    [self internalSetSymbology:SBSSymbologyEAN13 enabled:scanEAN13];
    [self internalSetSymbology:SBSSymbologyUPC12 enabled:scanEAN13];
}

- (BOOL)scanQRCODE {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyQR] enabled];
}

- (void)setScanQRCODE:(BOOL)scanQRCODE {
    [self internalSetSymbology:SBSSymbologyQR enabled:scanQRCODE];
}

- (BOOL)scanCODE128 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyCode128] enabled];
}

- (void)setScanCODE128:(BOOL)scanCODE128 {
    [self internalSetSymbology:SBSSymbologyCode128 enabled:scanCODE128];
}

- (BOOL)scanCODE39 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyCode39] enabled];
}

- (void)setScanCODE39:(BOOL)scanCODE39 {
    [self internalSetSymbology:SBSSymbologyCode39 enabled:scanCODE39];
}

- (BOOL)scanDATAMATRIX {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyDatamatrix] enabled];
}

- (void)setScanDATAMATRIX:(BOOL)scanDATAMATRIX {
    [self internalSetSymbology:SBSSymbologyDatamatrix enabled:scanDATAMATRIX];
}

- (BOOL)scanITF {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyITF] enabled];
}

- (void)setScanITF:(BOOL)scanITF {
    [self internalSetSymbology:SBSSymbologyITF enabled:scanITF];
}

- (BOOL)scanEAN5 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyFiveDigitAddOn] enabled];
}

- (void)setScanEAN5:(BOOL)scanEAN5 {
    [self internalSetSymbology:SBSSymbologyFiveDigitAddOn enabled:scanEAN5];
}

- (BOOL)scanEAN2 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyTwoDigitAddOn] enabled];
}

- (void)setScanEAN2:(BOOL)scanEAN2 {
    [self internalSetSymbology:SBSSymbologyTwoDigitAddOn enabled:scanEAN2];
}

- (BOOL)scanCODABAR {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyCodabar] enabled];
}

- (void)setScanCODABAR:(BOOL)scanCODABAR {
    [self internalSetSymbology:SBSSymbologyCodabar enabled:scanCODABAR];
}

- (BOOL)scanCODE93 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyCode93] enabled];
}

- (void)setScanCODE93:(BOOL)scanCODE93 {
    [self internalSetSymbology:SBSSymbologyCode93 enabled:scanCODE93];
}

- (BOOL)scanPDF417 {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyPDF417] enabled];
}

- (void)setScanPDF417:(BOOL)scanPDF417 {
    [self internalSetSymbology:SBSSymbologyPDF417 enabled:scanPDF417];
}

- (BOOL)scanGS1Databar {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyGS1Databar] enabled];
}

- (void)setScanGS1Databar:(BOOL)scanGS1Databar {
    [self internalSetSymbology:SBSSymbologyGS1Databar enabled:scanGS1Databar];
}

- (BOOL)scanGS1DatabarExpanded {
    return [[self.scanSettings settingsForSymbology:SBSSymbologyGS1DatabarExpanded] enabled];
}

- (void)setScanGS1DatabarExpanded:(BOOL)scanGS1DatabarExpanded {
    [self internalSetSymbology:SBSSymbologyGS1DatabarExpanded enabled:scanGS1DatabarExpanded];
}

- (void)internalSetSymbology:(SBSSymbology)symbology enabled:(BOOL)enabled {
    dispatch_async(self.settingsChangesQueue, ^{
        RedLaserSettingsChange *change = [[RedLaserSettingsChange alloc] init];
        change.symbology = symbology;
        change.enabled = enabled;
        [self.settingsChanges addObject:change];
        
        [self processSavedSettingsChanges];
    });
}

- (void)processSavedSettingsChanges {
    dispatch_async(self.settingsChangesQueue, ^{
        if (!self.applyingSettingsChanges && [self.settingsChanges count] > 0) {
            self.applyingSettingsChanges = YES;
            
            SBSScanSettings *settings = [self.scanSettings copy];
            
            for (RedLaserSettingsChange *savedChange in [self.settingsChanges copy]) {
                [settings setSymbology:savedChange.symbology enabled:savedChange.enabled];
                if (savedChange.symbology == SBSSymbologyTwoDigitAddOn
                        || savedChange.symbology == SBSSymbologyFiveDigitAddOn) {
                    settings.maxNumberOfCodesPerFrame = 2;
                }
            }
            [self.settingsChanges removeAllObjects];
            
            [self applyScanSettings:settings completionHandler:^{
                self.applyingSettingsChanges = NO;
                [self processSavedSettingsChanges];
            }];
        }
    });
}


#pragma mark - Other functions

- (BOOL)hasTorch {
    return YES;
}

- (void)turnTorch:(BOOL)value {
    [self switchTorchOn:value];
}

- (BOOL)isTorchOn {
    return NO;
}


#pragma mark - Unimplemented functions

- (bool)canFocus {
    return NO;
}

- (BOOL)isFocusing {
    return NO;
}

- (void)prepareToScan {}

- (void)clearResultsSet {}

- (void)startCollectingLocationData {}

- (void)requestCameraSnapshot:(bool) stillPictureSized {}

- (void)reportUnwantedBarcode:(BarcodeResult *) barcode {}

@end
